#ident "$Id: arena.c,v 1.11 2001/10/19 06:19:50 hpa Exp $"
/* ----------------------------------------------------------------------- *
 *   
 *   Copyright 2000-2001 H. Peter Anvin - All Rights Reserved
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as
 *   published by the Free Software Foundation, Inc.,
 *   59 Temple Place Ste 330, Boston MA 02111-1307, USA, version 2.1,
 *   incorporated herein by reference.
 *
 * ----------------------------------------------------------------------- */

/*
 * arena.c
 *
 * Persistent memory arena implemented using memory-mapping tricks
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h>
#include <inttypes.h>
#include <limits.h>
#include <fcntl.h>
#include <math.h>		/* HUGE_VAL */
#define __USE_MISC	1	/* Needed to support mremap() */
#define __USE_GNU	1	/* Needed to support mremap() */
#include <sys/mman.h>
#include <sys/stat.h>
#include <sched.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>

#include "lpsm.h"
#include "internals.h"

/* This might end up being a bitmask eventually */
enum page_status {
  page_clean = 0,
  page_dirty = 1,
};

/*
 * This is the data structure for the object store.  Note that only
 * one active object store is supported, due to the need to trap
 * SIGSEGV.
 */
struct ObjStore *lpsm_os_struct;

/* Wrappers for read() and write() which retries if incomplete */
static ssize_t lpsm_read(int fd, void *buf, size_t count)
{
  char *bufp = buf;
  ssize_t total = 0;
  ssize_t rv;

  while ( count ) {
    rv = read(fd, bufp, count);
    if ( rv == -1 ) {
      if ( errno == EINTR || errno == EAGAIN )
	continue;
      else
	return total ? total : -1;
    } else if ( rv == 0 ) {
      return total;
    }
    bufp  += rv;
    count -= rv;
    total += rv;
  }

  return total;
}

static ssize_t lpsm_write(int fd, void *buf, size_t count)
{
  char *bufp = buf;
  ssize_t total = 0;
  ssize_t rv;

  while ( count ) {
    rv = write(fd, bufp, count);
    if ( rv == -1 ) {
      if ( errno == EINTR || errno == EAGAIN )
	continue;
      else
	return total ? total : -1;
    } else if ( rv == 0 ) {
      return total;
    }
    bufp  += rv;
    count -= rv;
    total += rv;
  }

  return total;
}

/*
 * SIGSEGV handler for persistent object store
 */
static void lpsm_sigsegv(int signal, siginfo_t *siginfo, void *ptr)
{
  struct ObjStore *os = lpsm_os_struct;
  void *page;
  uintptr_t npage, offset;
  char *pageinfo;
  int old_errno = errno;
#ifdef __linux__
  struct sigcontext *ctxt;

# ifdef __i386__		/* This is so specific to Linux/i386 */
  /* Don't even begin to ask me where the 0x14 comes from */
  ctxt = (struct sigcontext *)((char *)ptr + 0x14);

  if ( siginfo->si_code == 0 ) {
    /* Old kernel.  Fill in data to the best of our knowledge. */
      if ( ctxt->trapno == 14 ) {
	/* Linux/i386 uses unmapped pages to mimic PROT_NONE, so we can't
	   tell ACCERR and MAPERR apart from the register state */
      siginfo->si_code = SEGV_ACCERR;
      siginfo->si_addr = (void *)ctxt->cr2;
    }
  }
# endif /* __i386__ */
#endif /* __linux__ */

  page = (void *)((uintptr_t)siginfo->si_addr & ~(os->pagesize-1));
  offset = (uintptr_t)page - (uintptr_t)os->arena;
  npage = (offset >> os->pageshift);
  pageinfo = os->pageinfo + npage;

  if ( signal != SIGSEGV || siginfo->si_code != SEGV_ACCERR ||
       offset >= os->arena_len ) {
    struct sigaction dfl;
    
    dfl.sa_handler = SIG_DFL;
    sigemptyset(&dfl.sa_mask);
    dfl.sa_flags = SA_ONESHOT;
    sigaction(SIGSEGV, &dfl, NULL);

#ifdef __linux
#ifdef __i386__
    /* Technically unsafe, but who cares at this point... */
    fprintf(stderr, "SIGSEGV at %p, %%eip = %p\n",
	    siginfo->si_addr, (void *)ctxt->eip);
#endif
#endif
    
    errno = old_errno;
    return;			/* Re-take fault */
  }
  
  mprotect(page, os->pagesize, PROT_READ|PROT_WRITE);

  switch ( (enum page_status) *pageinfo ) {
  case page_clean:
    *pageinfo = page_dirty;	/* Page now dirty */
    os->dirty_count++;		/* For accounting purposes */
    /* Leave page r/w */
    break;

  default:
    abort();			/* This shouldn't happen */
  }

  errno = old_errno;
}

/*
 * Routine to do log writeback.  Used by initial log recovery routine
 * as well as during-execution garbage collect.
 * THIS ROUTINE SHOULD BE INVOKED WITH LOCK HELD ON THE LOG FILE.
 */
static int lpsm_log_writeback(void)
{
  struct ObjStore *os = lpsm_os_struct;
  struct ObjStore_LogRecord record;
  off_t position, last_commit;
  struct flock lockmain;
  
  last_commit = 0;		/* Last COMMIT record found */
  position = lseek(os->log_fd, 0, SEEK_SET);
  
  while ( lpsm_read(os->log_fd, &record, sizeof(record)) == sizeof(record) ) {
    if ( record.magic != LOGRECORD_MAGIC )
      break;			/* Bad magic, assume rest of log corrupt */
    if ( record.record_type == osrec_commit ) {
      /* NOTE: last_commit points to the final byte to examine, thus
	 at the *end* of the final commit record. */
      position += sizeof(record);
      last_commit = position;	/* Found a commit record */
    } else if ( record.record_type == osrec_page ) {
      /* Advance past current page cluster */
      position = lseek(os->log_fd, record.size, SEEK_CUR);
    } else {
      return -1;		/* Unknown record - unsafe to process */
    }
  }
  
  /* Now we know where the last commit was.  Now we can process
     everything up to that point. */
  
  position = lseek(os->log_fd, 0, SEEK_SET);
  
  while ( lpsm_read(os->log_fd, &record, sizeof(record))
	  == sizeof(record) && position < last_commit ) {
    if ( record.magic != LOGRECORD_MAGIC )
      break;			/* Bad magic, assume rest of log corrupt */
    if ( record.record_type == osrec_commit ) {
      /* Found a commit record, do nothing */
      position += sizeof(record);
    } else if ( record.record_type == osrec_page ) {
      /* Write back data to file */
      char *data;
      
      position += sizeof(record);
      
      lockmain.l_type   = F_WRLCK;
      lockmain.l_whence = SEEK_SET;
      lockmain.l_start  = record.offset;
      lockmain.l_len    = record.size;
      while ( fcntl(os->main_fd, F_SETLKW, &lockmain) == -1 && errno == EINTR ); 
      data = mmap(NULL, record.size, PROT_WRITE, MAP_SHARED,
		  os->main_fd, record.offset);
      if ( data == MAP_FAILED )
	return -1;
      if ( lpsm_read(os->log_fd, data, record.size) != record.size )
	return -1;		/* Badness */
      if ( munmap(data, record.size) )
	return -1;
      
      lockmain.l_type = F_UNLCK;
      while ( fcntl(os->main_fd, F_SETLKW, &lockmain) == -1 && errno == EINTR );
      position += record.size;
    } else {
      return -1;		/* Unknown record - unsafe to process */
    }
  }
  
  /* Log successfully recovered.  Truncate. */
  fsync(os->main_fd);
  ftruncate(os->log_fd, 0);
  /* Write initial commit record, for sequence number recovery */
  record.magic       = LOGRECORD_MAGIC;
  record.record_type = osrec_commit;
  record.size        = os->fork_seq;
  record.offset      = 0x54494d43; /* For debugging */
  if ( lpsm_write(os->log_fd, &record, sizeof(record)) < sizeof(record) )
    return -1;

  fsync(os->log_fd);		/* Indicate log recovery complete */

  return 0;
}

/*
 * Routine to do log recovery
 */
static int lpsm_recover_log(void)
{
  struct ObjStore *os = lpsm_os_struct;
  struct flock lock;
  int rv = 0;

  /* First, lock the log file */
  lock.l_type   = F_WRLCK;
  lock.l_whence = SEEK_SET;
  lock.l_start  = 0;
  lock.l_len    = 0;
  while ( fcntl(os->log_fd, F_SETLKW, &lock) == -1 && errno == EINTR );

  /* Do log recovery, and write initial commit record. */
  rv = lpsm_log_writeback();

  /* Increase the sequence number, since we just wrote a commit. */
  os->fork_seq++;

  /* Unlock file and run. */
  lock.l_type = F_UNLCK;
  while ( fcntl(os->log_fd, F_SETLKW, &lock) == -1 && errno == EINTR );

  return rv;
}

/*
 * Opens the object store.  This includes log
 * playback (crash recovery) if the log file exists
 * and is nonempty.
 */
void *lpsm_arena_init(const char *main_file, const char *log_file, size_t *arena_len)
{
  struct ObjStore *os;
  void *arena_ptr;
  struct sigaction sigact;
  struct flock lock;
  off_t file_len, len = arena_len ? *arena_len : 0;
  size_t file_pages, len_pages;

  arena_ptr = ARENA_ADDRESS;

  lpsm_os_struct = os = malloc(sizeof(struct ObjStore));
  if ( !os )
    goto errx0;

  os->fork_seq = 0;		/* Initialize sequence counter */

  os->main_fd = open(main_file, O_RDWR|O_CREAT, 0666);
  if ( os->main_fd < 0 )
    goto errx1;

  os->pagesize = getpagesize();
  if ( os->pagesize & (os->pagesize - 1) )
    goto errx2;			/* WTF -- pagesize not a power of 2? */

  /* Compute log2(os->pagesize) */
  os->pageshift = 0;
  while ( (1 << os->pageshift) < os->pagesize )
    os->pageshift++;

  /*
   * Open log file
   */
  os->log_fd = open(log_file, O_RDWR|O_APPEND|O_CREAT, 0666);
  if ( os->log_fd < 0 )
    goto errx3;

  /* Now, do log recovery if needed */
  if ( lpsm_recover_log() )
    goto errx3;

  /* Allocate arena memory space */
  lock.l_type = F_WRLCK;
  lock.l_whence = SEEK_SET;
  lock.l_start  = 0;
  lock.l_len    = 0;
  while ( fcntl(os->main_fd, F_SETLKW, &lock) == -1 && errno == EINTR );
  file_len = lseek(os->main_fd, 0, SEEK_END);
  if ( len < file_len ) {
    len = file_len;
  }
  len = (len + os->pagesize - 1) & ~(os->pagesize - 1);
  if ( len > file_len ) {
    ftruncate(os->main_fd, len);	/* Extend file */
  }
  lock.l_type = F_UNLCK;
  while ( fcntl(os->main_fd, F_SETLKW, &lock) == -1 && errno == EINTR );

  os->arena = mmap(arena_ptr, len, PROT_READ, MAP_PRIVATE|MAP_FIXED, os->main_fd, 0);
  if ( os->arena == MAP_FAILED )
    goto errx3;

  os->arena_len = len;
  if ( *arena_len )
    *arena_len = len;

  os->pageinfo = malloc(len >> os->pageshift);
  if ( !os->pageinfo )
    goto errx4;

  file_len = (file_len + os->pagesize - 1) & ~(os->pagesize-1);
  file_pages = file_len >> os->pageshift;
  len_pages  = len >> os->pageshift;

  /* All pages clean at this time */
  memset(os->pageinfo, page_clean, len_pages);

  sigact.sa_sigaction = lpsm_sigsegv;
  sigemptyset(&sigact.sa_mask);
  sigact.sa_flags = SA_RESTART|SA_SIGINFO;
  if ( sigaction(SIGSEGV, &sigact, &os->oldact) )
    goto errx5;

  return os->arena;

 errx5:
  munmap(os->pageinfo, len >> os->pageshift);
 errx4:
  munmap(arena_ptr, len);
 errx3:
  if ( os->log_fd >= 0 ) close(os->log_fd);
 errx2:
  close(os->main_fd);
 errx1:
  free(os);
 errx0:

  return NULL;
}

/*
 * Object store checkpoint.  Writes entries to the log file.
 * The "gc_factor" is the factor of maximum log size file relative
 * to the arena size.  For example, if gc_factor == 0.5 then if the
 * log size is more than 50% of the arena file size a writeback cycle
 * will take place after the log has been written.  This means other
 * checkpoints will have to wait!
 *
 * Set gc_factor to 0.0 to force a gc, and to HUGE_VAL to inhibit gc.
 *
 * The "wait" argument determines the level of synchronicity required.
 *
 * Returns 0 on synchronous completion, 1 if the checkpoint was skipped,
 * (pid_t)-1 on error, or the pid_t of the asynchronous process.  Note:
 * if an asynchronous process is launched, and the next call uses
 * PSMSYNC_NONE, then it is up to the parent process to wait for the
 * launched processes; this process only remembers the last process launched.
 * FIXME: perhaps we should double-fork these processes?
 */
pid_t lpsm_checkpoint(double gc_factor, enum psmsync wait)
{
  static pid_t last_sync = 0;
  struct ObjStore *os = lpsm_os_struct;
  pid_t f, w;
  char *pi, *epi;
  void *page;
  
  pi  = os->pageinfo;
  epi = os->pageinfo + (os->arena_len >> os->pageshift);

  if ( last_sync ) {
    int status;
    switch (wait) {
    case PSMSYNC_NONE:
      break;
    case PSMSYNC_SKIP:
      w = waitpid(last_sync, &status, WNOHANG);
      if ( w == 0 )
	return 1;		/* Skip */
      if ( w == (pid_t)-1 ) {
	/* Note: under strace, waitpid() will return -1 instead of 0 */
	return (pid_t)-1;
      } else if ( WIFSIGNALED(status) || (WIFEXITED(status) && WEXITSTATUS(status)) ) {
	return (pid_t)-1;	/* Badness */
      }
      break;
    case PSMSYNC_WAIT:
    case PSMSYNC_SYNC:
      w = waitpid(last_sync, &status, 0);
      if ( w == (pid_t)-1 ) {
	if ( errno != ECHILD )
	  return (pid_t)-1;	/* Badness */
      } else if ( WIFSIGNALED(status) || (WIFEXITED(status) && WEXITSTATUS(status)) ) {
	return (pid_t)-1;	/* Badness */
      }
      break;
    }
  }

  f = fork();
  if ( f == (pid_t)-1 )
    return (pid_t)-1;		/* Checkpoint failed! */
  else if ( f > 0 ) {
    /* Parent process -- just mark all dirty pages clean */
    /* FIX: We probably want to do something more clever
       than memset() here, and perhaps keep around the old dirty
       array for a while. */
    mprotect(os->arena, os->arena_len, PROT_READ);
    memset(os->pageinfo, page_clean, os->arena_len >> os->pageshift);
    
    os->dirty_count = 0;	/* No pages dirty */
    os->fork_seq++;		/* Increase next sequence number */

    if ( wait == PSMSYNC_SYNC ) {
      int status;
      waitpid(f, &status, 0);
      last_sync = 0;
      if ( WIFSIGNALED(status) || (WIFEXITED(status) && WEXITSTATUS(status)) )
	return (pid_t)-1;	/* Something very bad happened... */
      else
	return 0;
    } else {
      last_sync = f;
      return f;
    }
  } else {
    /* Child process -- do the actual work of writing back dirty pages */

    struct ObjStore_LogRecord record, last_rec;
    struct flock lock;
    off_t logsize;

    record.magic = LOGRECORD_MAGIC;
    record.record_type = osrec_page;

    lock.l_whence = SEEK_SET;
    lock.l_start  = 0;
    lock.l_len    = 0;
    for (;;) {
      /* First, lock the entire log file */
      lock.l_type   = F_WRLCK;
      while ( fcntl(os->log_fd, F_SETLKW, &lock) == -1 && errno == EINTR );

      /* Make sure we were indeed next in turn */
      lseek(os->log_fd, -(off_t)sizeof(last_rec), SEEK_END);
      if ( lpsm_read(os->log_fd, &last_rec, sizeof(last_rec)) < sizeof(last_rec) ||
	   last_rec.magic != LOGRECORD_MAGIC ) {
	/* Something bad happened... */
	  kill(getppid(), SIGABRT); /* Kill main process */
	  _exit(99);
      }
      if ( last_rec.size+1 == os->fork_seq )
	break;			/* It's for us... */

      /* Someone else is ahead of us in line.  Yield to them. */
      lock.l_type = F_UNLCK;
      while ( fcntl(os->log_fd, F_SETLKW, &lock) == -1 && errno == EINTR );
      
      sched_yield();		/* Snore... */
    }

    /* Write dirty pages to log file */
    for ( pi = os->pageinfo ; pi < epi ; pi++ ) {
      if ( *pi == page_dirty ) {
	page = (char *)os->arena +
	  ((uintptr_t)(pi - os->pageinfo) << os->pageshift);
	record.offset = (char *)page - (char *)os->arena;

	/* Aggregate contiguous pages into a single record */
	record.size   = os->pagesize;
	while ( pi+1 < epi && pi[1] == page_dirty ) {
	  pi++;
	  record.size += os->pagesize;
	}

	if ( lpsm_write(os->log_fd, &record, sizeof(record))
	     < sizeof(record) ||
	     lpsm_write(os->log_fd, page, record.size) < record.size ) {
	  kill(getppid(), SIGABRT); /* Kill main process */
	  _exit(99);
	}
      }
    }

    /* This might be more efficiently done with fdatasync() */
    fsync(os->log_fd);		/* Make sure we have written everything */
    
    /* Write commit record */
    record.record_type = osrec_commit;
    record.size        = os->fork_seq;
    record.offset      = (off_t)0x54494d43;
    if ( lpsm_write(os->log_fd, &record, sizeof(record)) < sizeof(record) ) {
      kill(getppid(), SIGABRT);
      _exit(99);
    }
    fsync(os->log_fd);

    /* Check to see if it's time for garbage collect */
    logsize = lseek(os->log_fd, 0, SEEK_END);
    if ( gc_factor < HUGE_VAL && (double)logsize >= gc_factor*os->arena_len ) {
      /* Replaying the log isn't the most efficient way to do this.
	 We could also keep a status bit per page around, and flush
	 them out of the shadow array.  The biggest problem with that
	 is that it probably can't be done in the background, unlike
	 this method.  Leave this as-is for now. */
      if ( lpsm_log_writeback() ) {
	kill(getppid(), SIGABRT);
	_exit(99);
      }
    }

    /* Drop lock on log file */
    lock.l_type = F_UNLCK;
    while ( fcntl(os->log_fd, F_SETLKW, &lock) == -1 && errno == EINTR );

    _exit(0);			/* Done! */
  }
}

/*
 * Extend the size of the object store.
 *
 * This currently relies on several Linux-specific features,
 * specifically mremap() and knowing that we probably can extend
 * it without changing the virtual address.
 */
int lpsm_extend(size_t new_size)
{
  struct ObjStore *os = lpsm_os_struct;
  struct flock lock;
  void *newp, *infop;
  off_t file_size;
  int ft;
  size_t add_size, old_size;
  size_t add_pages, old_pages, new_pages, file_pages;
  struct rlimit fsizelimit;

  old_size = os->arena_len;

  if ( new_size <= old_size )
    return 0;			/* No action */

  new_size = (new_size + os->pagesize - 1) & ~(os->pagesize - 1);
  add_size = new_size - old_size;

  /* Make sure we won't run afoul of a set resource limit. */
  if ( getrlimit(RLIMIT_FSIZE, &fsizelimit) == 0 &&
       fsizelimit.rlim_cur != RLIM_INFINITY &&
       fsizelimit.rlim_cur < new_size )
    return -1;			/* Would exceed the file limit. */

  lock.l_type   = F_WRLCK;
  lock.l_whence = SEEK_SET;
  lock.l_start  = 0;
  lock.l_len    = 0;
  while ( fcntl(os->main_fd, F_SETLKW, &lock) == -1 && errno == EINTR );

  lock.l_type = F_UNLCK;
  file_size = lseek(os->main_fd, 0, SEEK_END);
  if ( file_size < new_size )
    ft = ftruncate(os->main_fd, new_size);
  else
    ft = 0;

  if ( ft )
    goto reset_size;		/* Failure */

  newp = mmap((char*)os->arena + old_size,
	      add_size, PROT_READ,
	      MAP_PRIVATE|MAP_FIXED, os->main_fd, old_size);
  
  if ( newp == MAP_FAILED )
    goto reset_size;		/* Failure */

  /* Since we specified MAP_FIXED, this should be guaranteed */
  assert( newp == (char*)os->arena + old_size );

  /* Convert sizes to pages */
  file_size  = (file_size + os->pagesize - 1) & ~(os->pagesize-1);
  new_pages  = new_size  >> os->pageshift;
  old_pages  = old_size  >> os->pageshift;
  file_pages = file_size >> os->pageshift;
  add_pages  = new_pages - old_pages;

  infop = realloc(os->pageinfo, new_pages);
  if ( !infop ) {
    munmap(newp, add_size);
    goto reset_size;
  }
  
  os->arena_len  = new_size;
  os->pageinfo   = infop;

  /* No more failure bailouts, unlock */
  while ( fcntl(os->main_fd, F_SETLKW, &lock) == -1 && errno == EINTR );

  /* Mark new pages clean */
  assert(new_pages >= file_pages);
  memset(os->pageinfo + file_pages, page_clean, new_pages-file_pages);

  return 0;

  /* We failed to extend the arena after ftruncate(), this is *usually*
     due to address space limitations. */
 reset_size:
  /* Restore the original file size */
  ftruncate(os->main_fd, file_size);
  /* Drop lock */
  while ( fcntl(os->main_fd, F_SETLKW, &lock) == -1 && errno == EINTR );
  return -1;
}

/*
 * Shut down the object store, free all resources.
 * THIS DOES NOT CHECKPOINT - call lpsm_checkpoint() first
 * if you want that functionality.  Calling this function without
 * first checkpointing and then calling lpsm_arena_init() can be used
 * to (very slowly) roll back to the last commit point.
 */
void lpsm_shutdown(void)
{
  struct ObjStore *os = lpsm_os_struct;

  munmap(os->arena, os->arena_len);
  free(os->pageinfo);
  sigaction(SIGSEGV, &os->oldact, NULL);
  close(os->log_fd);
  close(os->main_fd);
  free(os);
}
